/* gears.c */

/*
 * 3-D gear wheels.  This program is in the public domain.
 *
 * Brian Paul
 */

/* Conversion to GLUT by Mark J. Kilgard */

// ---- end of original demo info ----

// Modified by Ingemar, added FPS counter.
// OpenGL 3.2 version as of 13-02-05
// by Ingemar Ragnemalm
// No longer based on GLUT but on MicroGlut.
// Based on shaders and VAOs/VBOs as it should.
// 2022: Removed the "special" callback. It is not needed with MicroGlut. 

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "GL_utilities.h"
#include "VectorUtils3.h"
#include "MicroGlut.h"
// uses framework Cocoa
// uses framework OpenGL

#ifndef M_PI
#define M_PI 3.14159265
#endif

static long count = 0;
static long limit = -1;
static float startTime;

static GLfloat view_rotx = 20.0, view_roty = 30.0, view_rotz = 0.0;
static GLuint gear1, gear2, gear3; // Was display lists, now are VAOs
static GLfloat angle = 0.0;

int gear1Triangles=0, gear2Triangles=0, gear3Triangles=0;

GLuint program; // Shader
mat4 projectionMatrix;

static GLfloat pos[4] = { 5.0, 5.0, 10.0, 0.0 };
static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };


/*
 *  Draw a gear wheel and upload to a VAO.
 * 
 *  Input:  inner_radius - radius of hole at center
 *			 outer_radius - radius at center of teeth
 *			 width - width of gear
 *			 teeth - number of teeth
 *			 tooth_depth - depth of tooth
 
      Rewritten by Ingemar Ragnemalm 2013 to use VBOs/VAOs, GL3/4 friendly.
 */


static float *gVertices;
static float *gNormals;
static int numVertices;
static float currentNormalsX, currentNormalsY, currentNormalsZ;

static void AddVertex(float x, float y, float z)
{
	gVertices[numVertices*3] = x;
	gVertices[numVertices*3+1] = y;
	gVertices[numVertices*3+2] = z;
	gNormals[numVertices*3] = currentNormalsX;
	gNormals[numVertices*3+1] = currentNormalsY;
	gNormals[numVertices*3+2] = currentNormalsZ;
	numVertices += 1;
}

static void SetCurrentNormal(float x, float y, float z)
{
	currentNormalsX = x;
	currentNormalsY = y;
	currentNormalsZ = z;
}

static GLint gear(GLfloat inner_radius, GLfloat outer_radius, GLfloat width,
						GLint teeth, GLfloat tooth_depth, int *count)
{
	GLint i;
	GLfloat r0, r1, r2;
	GLfloat angle, da, da3;
	GLfloat u, v, len;
	
	gVertices = (float *) malloc(100000);
	gNormals = (float *) malloc(100000);
	numVertices = 0;
	
	r0 = inner_radius;
	r1 = outer_radius - tooth_depth / 2.0;
	r2 = outer_radius + tooth_depth / 2.0;
	
	da = 2.0 * M_PI / teeth / 4.0;
	
	SetCurrentNormal(0.0, 0.0, 1.0);

	/* draw front face */
	for (i = 0; i <= teeth; i++)
	{
		angle = i * 2.0 * M_PI / teeth;
		da3 = 4 * da;
		
		// step 1
		AddVertex(r0 * cos(angle), r0 * sin(angle), width * 0.5);
		AddVertex(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		AddVertex(r1 * cos(angle-da), r1 * sin(angle-da), width * 0.5);

		AddVertex(r0 * cos(angle), r0 * sin(angle), width * 0.5);
		AddVertex(r1 * cos(angle-da), r1 * sin(angle-da), width * 0.5);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), width * 0.5);

		// Step 2
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), width * 0.5);
		AddVertex(r1 * cos(angle-da), r1 * sin(angle-da), width * 0.5);
		AddVertex(r1 * cos(angle-da3), r1 * sin(angle-da3), width * 0.5);

		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), width * 0.5);
		AddVertex(r1 * cos(angle-da3), r1 * sin(angle-da3), width * 0.5);
		AddVertex(r0 * cos(angle-da3), r0 * sin(angle-da3), width * 0.5);
	}

	/* draw front sides of teeth */
	da = 2.0 * M_PI / teeth / 4.0;
	for (i = 0; i < teeth; i++)
	{
		angle = i * 2.0 * M_PI / teeth;

		AddVertex(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), width * 0.5);

		AddVertex(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), width * 0.5);
		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), width * 0.5);
	}

	SetCurrentNormal(0.0, 0.0, -1.0);
	da3 = 4*da;

	/* draw back face */
	for (i = 0; i <= teeth; i++)
	{
		angle = i * 2.0 * M_PI / teeth;
		
		AddVertex(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
		AddVertex(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), -width * 0.5);
		
		AddVertex(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), -width * 0.5);
		AddVertex(r1 * cos(angle-da), r1 * sin(angle-da), -width * 0.5);
		
		
		AddVertex(r1 * cos(angle-da), r1 * sin(angle-da), -width * 0.5);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), -width * 0.5);
		AddVertex(r0 * cos(angle-da3), r0 * sin(angle-da3), -width * 0.5);
		
		AddVertex(r1 * cos(angle-da), r1 * sin(angle-da), -width * 0.5);
		AddVertex(r0 * cos(angle-da3), r0 * sin(angle-da3), -width * 0.5);
		AddVertex(r1 * cos(angle-da3), r1 * sin(angle-da3), -width * 0.5);
	}

	/* draw back sides of teeth */
	da = 2.0 * M_PI / teeth / 4.0;
	for (i = 0; i < teeth; i++)
	{
		angle = i * 2.0 * M_PI / teeth;

		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), -width * 0.5);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);

		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
		AddVertex(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
	}

	SetCurrentNormal(0.0, 0.0, -1.0); // Copied from above

	/* draw outward faces of teeth */
	for (i = 0; i < teeth; i++)
	{
		angle = i * 2.0 * M_PI / teeth;

		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		u = r2 * cos(angle + da) - r1 * cos(angle);
		v = r2 * sin(angle + da) - r1 * sin(angle);
		len = sqrt(u * u + v * v);
		u /= len;
		v /= len;

// First quad
		AddVertex(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		AddVertex(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);

		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		AddVertex(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);

// Second quad
		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), -width * 0.5);

		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), width * 0.5);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), -width * 0.5);

// Third quad
		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), width * 0.5);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), -width * 0.5);
		u = r1 * cos(angle + 3 * da) - r2 * cos(angle + 2 * da);
		v = r1 * sin(angle + 3 * da) - r2 * sin(angle + 2 * da);
		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);

		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		AddVertex(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), width * 0.5);
		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), width * 0.5);
		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);

// Fourth quad
		u = r1 * cos(angle + 3 * da) - r2 * cos(angle + 2 * da);
		v = r1 * sin(angle + 3 * da) - r2 * sin(angle + 2 * da);
		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), width * 0.5);
		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);
		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		SetCurrentNormal(cos(angle+ 4 * da), sin(angle+ 4 * da), 0.0);
		AddVertex(r1 * cos(angle+ 4 * da), r1 * sin(angle+ 4 * da), -width * 0.5);

		u = r1 * cos(angle + 3 * da) - r2 * cos(angle + 2 * da);
		v = r1 * sin(angle + 3 * da) - r2 * sin(angle + 2 * da);
		SetCurrentNormal(v, -u, 0.0);
		AddVertex(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), width * 0.5);
		SetCurrentNormal(cos(angle), sin(angle), 0.0);
		SetCurrentNormal(cos(angle+ 4 * da), sin(angle+ 4 * da), 0.0);
		AddVertex(r1 * cos(angle+ 4 * da), r1 * sin(angle+ 4 * da), width * 0.5);
		AddVertex(r1 * cos(angle+ 4 * da), r1 * sin(angle+ 4 * da), -width * 0.5);

	}

	/* draw inside radius cylinder */
	da3 = 4*da;
	for (i = 0; i <= teeth; i++)
	{
		angle = i * 2.0 * M_PI / teeth;
		SetCurrentNormal(-cos(angle), -sin(angle), 0.0);
		AddVertex(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		AddVertex(r0 * cos(angle), r0 * sin(angle), width * 0.5);
		SetCurrentNormal(-cos(angle-da), -sin(angle-da), 0.0);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), width * 0.5);
		
		SetCurrentNormal(-cos(angle), -sin(angle), 0.0);
		AddVertex(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		SetCurrentNormal(-cos(angle-da), -sin(angle-da), 0.0);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), width * 0.5);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), -width * 0.5);
	
		
		SetCurrentNormal(-cos(angle-da), -sin(angle-da), 0.0);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), -width * 0.5);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), width * 0.5);
		SetCurrentNormal(-cos(angle-da3), -sin(angle-da3), 0.0);
		AddVertex(r0 * cos(angle-da3), r0 * sin(angle-da3), width * 0.5);
		
		SetCurrentNormal(-cos(angle-da), -sin(angle-da), 0.0);
		AddVertex(r0 * cos(angle-da), r0 * sin(angle-da), -width * 0.5);
		SetCurrentNormal(-cos(angle-da3), -sin(angle-da3), 0.0);
		AddVertex(r0 * cos(angle-da3), r0 * sin(angle-da3), width * 0.5);
		AddVertex(r0 * cos(angle-da3), r0 * sin(angle-da3), -width * 0.5);
	}
	
	// Build VAO and VBOs
	unsigned int vertexArrayObjID;
	unsigned int vertexBufferObjID;
	unsigned int normalsBufferObjID;
	// Allocate and activate Vertex Array Object
	glGenVertexArrays(1, &vertexArrayObjID);
	glBindVertexArray(vertexArrayObjID);
	// Allocate Vertex Buffer Objects
	glGenBuffers(1, &vertexBufferObjID);
	glGenBuffers(1, &normalsBufferObjID);
	
	// VBO for vertex data
	glBindBuffer(GL_ARRAY_BUFFER, vertexBufferObjID);
	glBufferData(GL_ARRAY_BUFFER, numVertices*3*sizeof(GLfloat), gVertices, GL_STATIC_DRAW);
	glVertexAttribPointer(glGetAttribLocation(program, "in_Position"), 3, GL_FLOAT, GL_FALSE, 0, 0); 
	glEnableVertexAttribArray(glGetAttribLocation(program, "in_Position"));
	
	// VBO for normals data
	glBindBuffer(GL_ARRAY_BUFFER, normalsBufferObjID);
	glBufferData(GL_ARRAY_BUFFER, numVertices*3*sizeof(GLfloat), gNormals, GL_STATIC_DRAW);
	glVertexAttribPointer(glGetAttribLocation(program, "in_Normal"), 3, GL_FLOAT, GL_FALSE, 0, 0); 
	glEnableVertexAttribArray(glGetAttribLocation(program, "in_Normal"));
	
	free(gVertices);
	free(gNormals);
	
	// Return VAO
	*count = numVertices;
	return vertexArrayObjID;
}

static void
draw(void)
{
//	mat4 rx, ry, rz, t, gearRot;
	mat4 globalR, m;
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
// GLOBAL ROTATION
	globalR = Mult(T(0, 0, -40.0), Mult(Rx(view_rotx*M_PI/180), Mult(Ry(-view_roty*M_PI/180), Rz(view_rotz*M_PI/180))));
	
// GEAR 1
	m = Mult(globalR, Mult(T(-3.0, -2.0, 0.0), Rz(angle*M_PI/180)));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, red);
	glBindVertexArray(gear1);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, gear1Triangles);

// GEAR 2
	m = Mult(globalR, Mult(T(3.1, -2.0, 0.0), Rz((-2.0 * angle - 9.0)*M_PI/180)));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, green);
	glBindVertexArray(gear2);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, gear2Triangles);

// GEAR 3
	m = Mult(globalR, Mult(T(-3.1, 4.2, 0.0), Rz((-2.0 * angle - 25.0)*M_PI/180)));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, blue);
	glBindVertexArray(gear3);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, gear3Triangles);
	
	glutSwapBuffers();
	
	count++;
	if (count == limit)
		exit(0);
	float theTime = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
	if (theTime >= startTime + 1.0)
	{
		printf("%ld fps\n", count);
		startTime = theTime;
		count = 0;
	}
}

static void
idle(void)
{
  angle += 2.0;
  glutPostRedisplay();
}

/* change view angle, exit upon ESC */
/* ARGSUSED1 */
static void
key(unsigned char k, int x, int y)
{
  switch (k)
  {
  case 'z':
    view_rotz += 5.0;
    break;
  case 'Z':
    view_rotz -= 5.0;
    break;
  case 27:  /* Escape */
    exit(0);
    break;
  case GLUT_KEY_UP:
    view_rotx += 5.0;
    break;
  case GLUT_KEY_DOWN:
    view_rotx -= 5.0;
    break;
  case GLUT_KEY_LEFT:
    view_roty += 5.0;
    break;
  case GLUT_KEY_RIGHT:
    view_roty -= 5.0;
  default:
    return;
  }
  glutPostRedisplay();
}

/* new window size or exposure */
static void
reshape(int width, int height)
{
	GLfloat h = (GLfloat) height / (GLfloat) width;

	glViewport(0, 0, (GLint) width, (GLint) height);
	projectionMatrix = frustum(-1.0, 1.0, -h, h, 5.0, 60.0);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
}


static void
init(void)
{
	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	
	program = loadShaders("gears.vert", "gears.frag");
	glUniform4fv(glGetUniformLocation(program, "pos"), 1, pos);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
	
	/* make the gears */
	gear1 = gear(1.0, 4.0, 1.0, 20, 0.7, &gear1Triangles);
	gear2 = gear(0.5, 2.0, 2.0, 10, 0.7, &gear2Triangles);
	gear3 = gear(1.3, 2.0, 0.5, 10, 0.7, &gear3Triangles);
	
	projectionMatrix = frustum(-1.0, 1.0, -1.0, 1.0, 5.0, 60.0);

	startTime = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
}

void visible(int vis)
{
  if (vis == GLUT_VISIBLE)
    glutIdleFunc(idle);
  else
    glutIdleFunc(NULL);
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	if (argc > 1) /* do 'n' frames then exit */
	   limit = atoi(argv[1]) + 1;
	else
		limit = 0;
	
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitContextVersion(3, 2);

	glutCreateWindow("GLUT Gears 3");
	init();

	glutDisplayFunc(draw);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(key);
	glutVisibilityFunc(visible);

	glutMainLoop();
	return 0;             /* ANSI C requires main to return int. */
}
